# -*- coding: utf-8 -*-
import os, glob
LXML = 1
try:
	from lxml.etree import iterparse
except ImportError:
	LXML ^= 1
	try:
		from xml.etree.cElementTree import iterparse
	except ImportError:
		from xml.etree.ElementTree import iterparse
from collections import defaultdict
from abc import abstractmethod
import gzip, zipfile, time
from .. import xml_unescape, isDreamOS
from ..config import E2m3u2bConfig
from ..modules.six import PY3, BytesIO
from ..modules.six.moves.urllib.parse import urlparse
from ..modules.six.moves import map
try:
	from backports import lzma
except ImportError:
	try:
		import lzma
	except ImportError:
		import subprocess
		try:
			subprocess.call(["apt" if isDreamOS else "opkg", "install", "python3-lzma" if PY3 else "python-lzma"])
			import lzma
		except:
			subprocess.call(["apt" if isDreamOS else "opkg", "install", "python3-backports-lzma" if PY3 else "python-backports-lzma"])
			from backports import lzma

from Components.config import config

histseconds = 10800  # 3 hours ago
timespan = int(time.time() + 86400 * 7)  # 7 days ahead
try:
	histseconds = int(config.epg.histminutes.value) * 60
except:
	if isDreamOS:
		try:
			histseconds = int(config.misc.epgcache_outdated_timespan.value) * 3600
			timespan = int(time.time() + int(config.misc.epgcache_timespan.value) * 86400)
		except:
			pass

logger = E2m3u2bConfig.logger
channelCache = {}


def isLocalFile(filename):
	url_parsed = urlparse(filename)
	if url_parsed.scheme in ('file', ''): # Possibly a local file
		return os.path.exists(url_parsed.path)
	return False

def getChannels(path, name=None):
	global channelCache
	if name in channelCache:
		return channelCache[name]
	dirname, filename = os.path.split(path)
	if name:
		channelfile = os.path.join(dirname, name) if isLocalFile(name) else name
	else:
		channelfile = os.path.join(dirname, filename.split('.', 1)[0] + '.channels.xml')
	try:
		return channelCache[channelfile]
	except KeyError:
		channelCache[channelfile] = EPGChannel(channelfile)
		return channelCache[channelfile]

def getFiletype(fp):
	"""
	Attempts to detect file or file like object format from the content by reading the first 10 bytes.
	Returns None if no format could be detected.
	https://en.wikipedia.org/wiki/List_of_file_signatures
	"""
	if hasattr(fp, 'seek'):
		bs = fp.read(10)
		fp.seek(0, 0)
	else:
		with open(fp, "rb") as f:
			bs = f.read(10)

	if bs[:2] == b"\x1f\x8b":
		return "GZ"
	elif bs[:6] == b"\xfd\x37\x7a\x58\x5a\x00":
		return "XZ"
	elif bs[:4] == b"\x50\x4b\x03\x04":
		return "ZIP"
	elif bs[:6] == b"\x3c\x3f\x78\x6d\x6c\x20":
		return "XML"
	elif bs[:9] == b"\xef\xbb\xbf\x3c\x3f\x78\x6d\x6c\x20":
		return "XML"  # utf-8-bom
	return None

def getDoc(fp):
	# get an iterable
	doc = iterparse(fp, ('start', 'end'), encoding='utf-8', huge_tree=True) if LXML else iterparse(fp, events=('start', 'end'))
        # get the root element
	_, root = next(doc)
	return doc, root

def enumerateXML(fp, tag=None):
	"""Enumerates ElementTree nodes from file object 'fp'"""
	if tag is None:
		logger.warn("[enumerateXML] There is nothing no enumerate. Node name required")
		return
	d = {'supplier': 'providers', 'programme': 'events', 'channel': 'channels',
		'category': 'categories', 'groupe': 'categories of channels', 'group': 'additional xmltv sources'}
	doc, root = getDoc(fp)
	depth = intermediate = 0

	with Timer() as timer:
		for event, element in doc:
			if element.tag == tag:
				depth += [-1,1][event == 'start']
				if not depth:
					yield element
					if LXML:
						while element.getprevious() is not None:
							del element.getparent()[0]
					element.clear()
					timer.counter += 1

			ctime = time.time()
			if ctime - timer.start >= 10:
				if timer.counter > intermediate:
					logger.info("Processed: %s %s" % (timer.counter, d.get(tag, 'records')))
					intermediate = timer.counter
					root.clear()
				else:
					logger.info("Iterating over XML nodes ...")
				timer.start = ctime

		timer.msg = 'Total processed {} {}'.format(timer.counter, d.get(tag, 'records'))

def openStream(filename):
	try:
		if not os.path.getsize(filename):
			raise Exception("file is empty: %s" % filename)
	except:
		raise Exception("unexpected error occurred with file: %s" % filename)

	ftype = getFiletype(filename)
	try:
		if ftype == 'GZ':
			fd = gzip.open(filename, 'rb')
		elif ftype == 'XZ':
			fd = lzma.open(filename, 'rb')
		elif ftype == 'ZIP':
			zip_obj = zipfile.ZipFile(filename, 'r')
			fd = BytesIO(zip_obj.open(zip_obj.namelist()[0]).read())
		elif ftype == 'XML':
			fd = open(filename, 'rb')
		else:
			raise KeyError(ftype)
	except KeyError:
		raise("unsupported file type: %s" % filename)
	except Exception:
		raise Exception("file %s is not a valid %s" % (filename, ftype))

	if getFiletype(fd) != 'XML':
		raise Exception("file %s is not a valid XML" % filename)
	else:
		return fd


class Timer(object):
	def __init__(self):
		self._msg = ''

	def __enter__(self):
		self._start = time.time()
		self._counter = 0
		return self

	def __exit__(self, *args):
		logger.info(self.msg)
		self.close()

	def __del__(self):
		try:
			self.close()
		except Exception:
			pass
	@property
	def start(self):
		return self._start

	@start.setter
	def start(self, start):
		self._start = start

	@property
	def elapsed(self):
		return time.time() - self.start

	@property
	def msg(self):
		return self._msg

	@msg.setter
	def msg(self, msg):
		self._msg = msg

	@property
	def counter(self):
		return self._counter

	@counter.setter
	def counter(self, cnt):
		self._counter = cnt

	@abstractmethod
	def close(self):
		"""Called when exiting the context manager"""


class EPGChannel(object):
	def __init__(self, filename, urls=None):
		self.mtime = None
		self.name = filename
		self.urls = [filename] if urls is None else urls
		self.items = defaultdict(set)

	def parse(self, filterCallback, downloadedFile):
		''' parse  *****_cahnnels.xml lookup table file '''
		logger.info("Enumerating channels lookup table XML-file")
		self.items = defaultdict(set) # set make the reference unique to avoid loading twice the same EPG data.
		try:
			for elem in enumerateXML(openStream(downloadedFile), 'channel'):
				id, ref = list(map(xml_unescape, [elem.get('id', '').lower(), elem.text]))
				if id and ref and filterCallback(ref):
					self.items[id].add(':'.join(ref.split(':')[:10]))
		except Exception as e:
			logger.error("EPGConfig failed to parse %s Error: %s" % (self.name, e))
		else:
			logger.info("Formed lookup table for: %s channels id" % len(self.items))

	def update(self, filterCallback, downloadedFile=None):
		if downloadedFile is not None:
			self.mtime = time.time()
			self.parse(filterCallback, downloadedFile)
		elif len(self.urls) == 1 and isLocalFile(self.urls[0]):
			mtime = os.path.getmtime(self.urls[0])
			if (not self.mtime) or (self.mtime < mtime):
				self.mtime = mtime
				self.parse(filterCallback, self.urls[0])

	def downloadables(self):
		if not (len(self.urls) == 1 and isLocalFile(self.urls[0])):
			# Check at most once a day
			if (not self.mtime) or (time.time() - self.mtime <= 86400):
				return self.urls
		return []

	def __repr__(self):
		return "EPGChannel(urls=%s, channels=%s, mtime=%s)" % (self.urls, self.items and len(self.items), self.mtime)


class EPGSource(object):
	def __init__(self, sourcefile, elem, category=None):
		self.parser = elem.get('type', 'gen_xmltv')
		self.nocheck = int(elem.get('nocheck', 0))
		self.urls = [ xml_unescape(e.text) for e in elem.findall('url') ]
		self.url = self.urls.pop()
		self.description = xml_unescape(elem.findtext('description', self.url))
		self.category = category
		self.channels = getChannels(sourcefile, elem.get('channels'))


def enumSourcesFile(sourcefile, filter=None, categories=False):
	''' parse  /etc/epgimport/*****.sources.xml|.gz|.xz|.zip '''
	global channelCache
	category = None
	try:
		doc, root = getDoc(openStream(sourcefile))
		for event, elem in doc:
			if event == 'start':
				# Need the category name sooner than the contents, hence "start"
				if elem.tag == 'sourcecat':
					category = xml_unescape(elem.get('sourcecatname'))
					if categories:
						yield category

			elif event == 'end':
				if elem.tag == 'source':
					s = EPGSource(sourcefile, elem, category)
					if (filter is None) or (s.description in filter):
						yield s
					elem.clear()
				elif elem.tag == 'channel':
					name = xml_unescape(elem.get('name'))
					urls = [ xml_unescape(e.text) for e in elem.findall('url') ]
					try:
						channelCache[name].urls = urls
					except:
						channelCache[name] = EPGChannel(name, urls)
					elem.clear()
				elif elem.tag == 'sourcecat':
					category = None

				root.clear()

	except Exception as e:
		logger.error("EPGConfig enumSourcesFile Error: %s" % e)

def enumSources(path, filter=None, categories=False):
	try:
		for file in glob.iglob(os.path.join(path, '*.sources.*')):
			for s in enumSourcesFile(file, filter, categories):
				yield s

	except Exception as e:
		logger.error("EPGConfig enumSources Error: %s" % e)
