# -*- coding: utf-8 -*-
import os, errno, time, emoji
from tempfile import gettempdir
from .. import isDreamOS, medialist
from ..modules.six import BytesIO
from ..config import E2m3u2bConfig
from .EPGConfig import isLocalFile, openStream, Timer
from twisted.internet import reactor, threads, ssl, _sslverify, defer, fdesc
from twisted.web.client import Agent, BrowserLikeRedirectAgent, readBody, HTTPConnectionPool
try:
	from twisted.web.client import BrowserLikePolicyForHTTPS as IgnoreHTTPSContextFactory
except:
	# for twisted < 14 backward compatibility
	from twisted.web.client import WebClientContextFactory as IgnoreHTTPSContextFactory
from twisted.internet.defer import setDebugging
from twisted.web.http_headers import Headers
from twisted.python.runtime import platform
from collections import namedtuple
try:
	from backports import lzma
except ImportError:
	import lzma

class IgnoreHTTPS(IgnoreHTTPSContextFactory):
	def creatorForNetloc(self, hostname, port):
		options = ssl.CertificateOptions(verify=False)
		return _sslverify.ClientTLSOptions(hostname.encode('utf-8'), options.getContext())

setDebugging(False)
logger = E2m3u2bConfig.logger
downloadTimeout = int(E2m3u2bConfig.downloadTimeout) if (10 <= E2m3u2bConfig.downloadTimeout <= 300) else 60
ISO639_associations = {}
headers = Headers({'User-Agent': ['Mozilla/5.0 (SmartHub; SMART-TV; U; Linux/SmartTV; Maple2012) AppleWebKit/534.7 (KHTML, like Gecko) SmartTV Safari/534.7']})
# Use a persistent connection pool.
pool = HTTPConnectionPool(reactor, persistent=True)
pool.maxPersistentPerHost = 1
pool.cachedConnectionTimeout = 600
agent = BrowserLikeRedirectAgent(Agent(reactor, pool=pool, contextFactory=IgnoreHTTPS()))
PARSERS = {
		'xmltv': ['xmltvconverter', 'XMLTVConverter'],
		'gen_xmltv': ['xmltvconverter', 'XMLTVConverter'],
		'jtv': ['jtvconverter', 'JTVConverter'],
		'gen_jtv': ['jtvconverter', 'JTVConverter'],
	}

def completed(passthrough):
	if timeoutCall.active():
		timeoutCall.cancel()
	return passthrough

def doRead(response):
	d = readBody(response)
	d.addCallback(getISO639)

def getISO639(text):
	for l in text.decode('UTF-8').split('\n'):
		l = l.split('|')
		if l and l[2]: ISO639_associations[l[2]] = l[0]

d = agent.request(b'GET', b'http://loc.gov/standards/iso639-2/ISO-639-2_utf-8.txt', headers, None)
d.addCallback(doRead).addErrback(lambda e: logger.warn("Failed to load ISO639 codes: %s" % e))
#set a timeout for all request. If unsuccessfull then call the errback
timeoutCall = reactor.callLater(downloadTimeout, d.cancel)
d.addBoth(completed)

def getParser(name):
	name, cls = PARSERS.get(name)
	return getattr(__import__(__name__[:__name__.rfind('.')+1] + name, fromlist=[None]), cls)

def bigStorage(minFree, default=gettempdir(), *candidates):

	def free_space_available(dirname):
		try:
			st = os.statvfs(dirname)
			return st.f_bavail * st.f_frsize
		except:
			return 0

	candidates = list(filter(os.path.isdir, candidates))
	candidates.append(default)
	try:
		candidate = sorted(candidates, key=lambda x: free_space_available(x), reverse=True)[0]
		free = free_space_available(candidate)
		if (free > minFree) or (free > free_space_available(default)):
			return candidate
		raise Exception("Free space is critically small")
	except Exception as e:
		logger.error("EPGImport an error occurred in bigStorage detection: %s" % repr(e))
		return default

def unlink_if_exists(filename):
	try:
		if filename is not None:
			os.unlink(filename)
	except OSError as e:
		if e.errno != errno.ENOENT: # errno.ENOENT = no such file or directory
			raise

def human_size(bytes, units=[' bytes','KB','MB','GB','TB','PB','EB']):
	'''a human readable string representation of bytes'''
	return str(bytes) + units[0] if bytes < 1024 else human_size(bytes>>10, units[1:])


class OudeisImporter(object):
	'Wrapper to convert original patch to new one that accepts multiple services'

	def __init__(self, epgcache):
		self.epgcache = epgcache
	# difference with old patch is that services is a list or tuple, this
	# wrapper works around it.

	def importEvents(self, services, events):
		for service in services:
			self.epgcache.importEvent(service, events)

class EPGImport(object):
	"""Simple Class to import EPGData"""

	HDD_EPG_DAT = '/etc/enigma2/epg.db' if isDreamOS else '/etc/enigma2/epg.dat'
	tempfile = os.path.join(bigStorage(9000000, *medialist), 'epgimport.tmp')

	def __init__(self, epgcache, channelFilter):
		self.timer = Timer()
		self.storage = None
		self.sources = []
		self.source = None
		self.fd = None
		self.iterator = None
		self.onDone = None
		self.epgcache = epgcache
		self.channelFilter = channelFilter

	def cacheStateChanged(self, state):
		'''  0 = started, 1 = stopped, 2 = aborted, 3 = deferred, 4 = load_finished, 5 = save_finished '''
		if state.state == 5:
			logger.debug("EPGCache save finished")
			import epg_importer
			self.storage = epg_importer.EPGClass(self.HDD_EPG_DAT)
			self.nextImport()

		elif state.state == 4:
			if isDreamOS:
				logger.debug("EPGCache load finished")
			self.storage = None
			self.source = None
			if self.onDone:
				self.onDone(reboot=False, epgfile=None)

	def saveEPGCache(self):
		if isDreamOS:
			self.cacheState_conn = self.epgcache.cacheState.connect(self.cacheStateChanged)
		logger.info("Save the EPGchache to database file %s ..." % self.HDD_EPG_DAT)
		self.epgcache.save()

	def beginImport(self):
		'''Starts importing using Enigma reactor'''
		self.timer.counter = 0
		self.timer.start = time.time()

		attr = dir(self.epgcache)
		if not all(p in attr for p in ('save', 'load')):
			logger.error("The image being used does not have a eEPGCache save() or load() methods, EPG import is not possible")
			self.closeImport()
			return

		if isDreamOS:
			logger.info("DreamOS detected...")
			self.saveEPGCache()
			return
		elif 'importEvents' in attr:
			self.storage = self.epgcache
		elif 'importEvent' in attr:
			self.storage = OudeisImporter(self.epgcache)
		else:
			logger.error("The image being used does not have any valid eEPGCache methods, EPG import is not possible")
			self.closeImport()
			return
		self.nextImport()

	def fetchUrl(self, filelink, cb, errb):
		d = self.do_download(filelink)
		d.addCallback(cb).addErrback(errb)

	def createIterator(self, filename):
		self.source.channels.update(self.channelFilter, filename)
		return getParser(self.source.parser).enumFile(self.fd, self.source.channels.items)

	def nextImport(self):
		if self.fd is not None:
			self.closeReader()
		try:
			self.source = self.sources.pop()
		except:
			self.closeImport()
		else:
			logger.info('### required downloads started, source = "%s" %s' % (self.source.description, emoji.emojize(':open_file_folder:')))
			self.fetchUrl(self.source.url, self.afterDownload, self.downloadFail)

	def afterDownload(self, result):
		logger.debug("afterDownload started")
		try:
			if 'jtv' in self.source.parser:
				with open(self.tempfile, "rb") as f:
					self.fd = BytesIO(f.read())
			else:
				self.fd = openStream(self.tempfile)

		except Exception as e:
			self.downloadFail(e)
			return
		finally:
			unlink_if_exists(self.tempfile)

		self.channelFiles = self.source.channels.downloadables()
		try:
			filename = self.channelFiles.pop()
			self.fetchUrl(filename, lambda x: self.afterChannelDownload(x, self.tempfile), self.channelDownloadFail)
		except IndexError:
			self.afterChannelDownload(None, None)

	def afterChannelDownload(self, result, filename):
		logger.debug("afterChannelDownload started")
		self.iterator = self.createIterator(filename)
		unlink_if_exists(filename)
		logger.info('### import EPG events started, source = "%s" %s' % (self.source.description, emoji.emojize(':sign_of_the_horns:')))
		if platform.supportsThreads():
			logger.info("### using twisted thread, yay! %s" % emoji.emojize(':rocket:'))
			threads.deferToThread(self.doThreadRead).addErrback(lambda ignore: self.nextImport())
		else:
			logger.info("### using twisted reactor, not bad! %s" % emoji.emojize(':racing_car:'))
			self.start()

	def importEvents(self):
		data = next(self.iterator)
		if data:
			try:
				self.storage.importEvents(*data)
				self.timer.counter += len(data[1])
			except Exception as e:
				logger.error("EPGImport ### importEvents exception: %s" % repr(e))

	def doThreadRead(self):
		'''This is used on images with threading'''
		while 1:
			self.importEvents()

	def doRead(self):
		'''called from reactor to read data'''
		try:
			self.importEvents()
		except StopIteration:
			self.stop()

	def fileno(self):
		'''called from reactor'''
		return self.fd.fileno() if self.fd else None

	def connectionLost(self, failure):
		'''called from reactor on lost connection'''
		# This happens because enigma calls us after removeReader
		pass

	def logPrefix(self):
		'''called from reactor'''
		return '[EPGImport]'

	def start(self):
		reactor.addReader(self)

	def stop(self):
		reactor.removeReader(self)
		self.nextImport()

	def channelDownloadFail(self, failure):
		logger.info("Download channels lookup table failed: %s" % failure)
		try:
			filename = self.channelFiles.pop()
			logger.info("Attempting alternative XML-file's for channels lookup table")
			self.fetchUrl(filename, self.afterChannelDownload, self.channelDownloadFail)
		except IndexError:
			logger.info("No more alternatives lookup table XML-file's for channels")
			self.nextImport()

	def downloadFail(self, failure):
		logger.warn("Download EPG failed: %s" % failure)
		try:
			self.source.url = self.source.urls.pop()
			logger.info("Attempting alternative url's for EPG")
			self.fetchUrl(self.source.url, self.afterDownload, self.downloadFail)
		except IndexError:
			logger.info("No more alternatives url's for EPG")
			self.nextImport()

	def closeReader(self):
		try:
			self.fd.flush()
			os.fsync(self.fd.fileno())
			self.fd.close()
		except:
			pass
		self.fd = None
		self.iterator = None
		logger.info('### import events finished, source = "%s" %s' % (self.source.description, emoji.emojize(':thumbs_up:')))

	def closeImport(self):
		if self.timer.counter:
			if isDreamOS:
				logger.info("Load the database file %s to EPGCache ..." % self.HDD_EPG_DAT)
				self.epgcache.load()
			else:
				self.saveEPGCache()

			self.timer.msg = 'Total imported {} events in {}'.format(self.timer.counter, time.strftime("%H:%M:%S", time.gmtime(self.timer.elapsed)))
		else:
			logger.info('#### No events to import! #### %s' % emoji.emojize(':pouting_face:'))

		self.timer.__exit__()
		if not isDreamOS:
			self.cacheStateChanged(namedtuple('state', 'state')(4))

	@property
	def eventCount(self):
		return self.timer.counter

	@property
	def isImportRunning(self):
		return self.source is not None

	@defer.inlineCallbacks
	def do_download(self, filelink):
		global timeoutCall
		scheme = filelink.split('/')[0]
		if not(scheme in ('http:', 'https:', 'file:') or isLocalFile(filelink)):
			raise ValueError("Unsupported scheme: %s" % scheme)

		logger.info("Downloading: %s " % filelink)
		d = agent.request(b'GET', filelink.encode('utf-8'), headers, None)
		d.addCallback(readBody)
		timeoutCall = reactor.callLater(downloadTimeout, d.cancel)
		d.addBoth(completed)

		data = yield d
		with open(self.tempfile, 'wb') as fd:
			d = fdesc.writeToFD(fd.fileno(), data)
			logger.info('%s downloaded' % human_size(d))
