# -*- coding: utf-8 -*-
import os, time
from ..config import E2m3u2bConfig
from ctypes import c_int32
from sqlite3 import dbapi2 as sqlite
from enigma import eEPGCache


logger = E2m3u2bConfig.logger
# For SQLite INTEGER type compatibility with unit32_t
class Hash(object):

	def __init__(self, s):
		self.hashedstring = s.encode('utf-8').strip() if s else ''

	def __conform__(self, protokol):
		if protokol is sqlite.PrepareProtocol:
			return c_int32(eEPGCache.getStringHash(self.hashedstring)).value


class epgdb_class(object):

	events = []

	def __init__(self, source_name, source_priority, epgdb_file):
		self.source_name = source_name
		self.priority = source_priority
		self.epgdb = epgdb_file

		try:
			self.connection = sqlite.connect(self.epgdb, timeout=20, isolation_level='DEFERRED', check_same_thread=False)
			self.connection.create_function("int32", 1, lambda d: c_int32(d).value)
			self.connection.text_factory = lambda x: x.encode('utf-8')
			self.cursor = self.connection.cursor()
			self.cursor.executescript("""
CREATE INDEX IF NOT EXISTS unique_title_hash ON T_Title (hash);
CREATE INDEX IF NOT EXISTS unique_event_idx ON T_Event (service_id, dvb_event_id);
""")
			self.cursor.execute("SELECT id from T_Source WHERE source_name=? and priority=?", (self.source_name, self.priority))
			row = self.cursor.fetchone()
			if row:
				self.source_id = row[0]
			else:
				with self.connection:
					self.cursor.execute("INSERT INTO T_Source (source_name, priority) VALUES (?,?)", (self.source_name, self.priority))
					self.source_id = self.cursor.lastrowid

			logger.debug("[EPGDB] connect to {epgdb} finished".format(**self.__dict__))
		except:
			raise Exception("[EPGDB] connect to {epgdb} failed".format(**self.__dict__))

	def add_event(self, event):
		# (starttime, duration, title, sort_descr, extended_descr, language)
		self.events.append(event)

	def description_id(self, table, descr):
		descr_hash = Hash(descr)
		self.cursor.execute("SELECT id from %s_Description WHERE hash=?" % table, (descr_hash,))
		row = self.cursor.fetchone()
		if row:
			return row[0]
		else:
			cmd = "INSERT INTO %s_Description  (hash, %s_description) VALUES(?,?)" % (table, table[2:].lower())
			self.cursor.execute(cmd, (descr_hash, descr))
			return self.cursor.lastrowid

	def title_id(self, table, title):
		title_hash = Hash(title)
		self.cursor.execute("SELECT id from %s WHERE hash=?" % table, (title_hash,))
		row = self.cursor.fetchone()
		if row:
			return row[0]
		else:
			self.cursor.execute("INSERT INTO %s (hash, title) VALUES(?,?)" % table, (title_hash, title))
			return self.cursor.lastrowid

	def preprocess_events_channel(self, services):
		if not self.events:
			return
		with self.connection:
			for service in services:
				# sid:tsid:onid:dvbnamespace
				#{:04x}:{:04x}:{:04x}:{:08x}
				sref = tuple(int(x, 16) for x in service.split(":")[3:7])
				self.cursor.execute("SELECT id from T_Service WHERE sid=? and tsid=? and onid=? and dvbnamespace=int32(?)", sref)
				service = self.cursor.fetchone()
				if service:
					service_id = service[0]
				else:
					self.cursor.execute("INSERT INTO T_Service (sid, tsid, onid, dvbnamespace) VALUES(?,?,?,int32(?))", sref)
					service_id = self.cursor.lastrowid

				for event in self.events:
					begin_time, duration = event[0:2]
					dvb_event_id = Hash("%d%s" % (begin_time, event[5]))

					self.cursor.execute("SELECT id FROM T_Event WHERE service_id=? and dvb_event_id=?", (service_id, dvb_event_id))
					row = self.cursor.fetchone()
					if row:
						event_id = row[0]
					else:
						cmd = "INSERT INTO T_Event (service_id, begin_time, duration, source_id, dvb_event_id) VALUES(?,?,?,?,?)"
						self.cursor.execute(cmd, (service_id, begin_time, duration, self.source_id, dvb_event_id))
						event_id = self.cursor.lastrowid
						title_id = self.title_id('T_Title', (u"%s" % event[2])[:240])

					short_d, long_d = event[3:5]
					short_description_id = self.description_id('T_Short', short_d) if short_d else None
					extended_description_id = self.description_id('T_Extended', long_d) if long_d else None
					if row:
						# Update data for an existing event
                                        	cmd = "UPDATE T_Data SET short_description_id=?, extended_description_id=? WHERE ROWID=?"
                                        	self.cursor.execute(cmd, (short_description_id, extended_description_id, event_id))
					else:
						# Create new event
						cmd = "INSERT INTO T_Data (event_id, title_id, short_description_id, extended_description_id, iso_639_language_code) VALUES(?,?,?,?,?)"
						self.cursor.execute(cmd, (event_id, title_id, short_description_id, extended_description_id, event[5]))

			self.events = []

	def final_process(self):
		self.cursor.executescript("""
DROP INDEX IF EXISTS unique_title_hash;
DROP INDEX IF EXISTS unique_event_idx;
PRAGMA incremental_vacuum;
PRAGMA analysis_limit=400;
PRAGMA optimize;
""")
		logger.debug("[EPGDB] Total changes made in {epgdb}: {connection.total_changes}".format(**self.__dict__))
		self.cursor.close()
		self.connection.close()
		logger.debug("[EPGDB] Finished generating the SQL file {epgdb}".format(**self.__dict__))
