import re, datetime, time, struct, zipfile
from datetime import datetime, timedelta
from collections import deque
from .EPGConfig import Timer, histseconds, timespan
from .EPGImport import ISO639_associations
from .. import isDreamOS
from ..config import E2m3u2bConfig
from ..modules.six import PY3, ensure_str, ensure_text
from ..modules.six.moves import range, zip, filter

logger = E2m3u2bConfig.logger
PATTERN = re.compile(r'''\.(?=(?:[^'"]|'[^']*'|"[^"]*")*$)''')
LANG = ISO639_associations.get(language.getLanguage()[:2], 'eng')
EPOCH_AS_FILETIME = 116444736000000000  # January 1, 1970 as MS file time
HUNDREDS_OF_NANOSECONDS = 10000000

def utc_offset():
	timestamp = time.time()
	utime = datetime.utcfromtimestamp(timestamp)
	btime = datetime.fromtimestamp(timestamp)
	return (btime - utime).total_seconds()

utc_offset = utc_offset()

def MSfiletime_to_timestamp(MStimestamp):
	timestamp, _ = divmod((MStimestamp - EPOCH_AS_FILETIME), HUNDREDS_OF_NANOSECONDS)
	return int(timestamp - utc_offset)

length = lambda d: struct.unpack('<H', d[:2])[0]

def parse_events(data):
	pack_len = 12
	events_num = length(data) * pack_len
	data = data[2:]
	return [ MSfiletime_to_timestamp(struct.unpack('<hqH', data[i:i+pack_len])[1]) for i in range(0, events_num, pack_len) ]

def parse_titles(data, encoding):
	if data[:26] != b'\x4a\x54\x56\x20\x33\x2e\x78\x20\x54\x56\x20\x50\x72\x6f\x67\x72\x61\x6d\x20\x44\x61\x74\x61\x0a\x0a\x0a':
		raise Exception('Invalid JTV 3.x TV Programm Data format')
	data = data[26:]
	while data:
		title_length = length(data)
		data = data[2:]
		yield data[:title_length].decode(encoding)   # yield title
		data = data[title_length:]


class JTVConverter(object):

	@classmethod
	def enumFile(cls, jtv_filename, channelsDict={}, jtv_encoding="cp1251"):
		if not channelsDict:
			logger.warn("JTVConverter is nothing to enumerate. The channels lookup table is empty")
			return
		now_timestamp_utc = int(time.time()) - histseconds
		logger.info("Keep outdated EPG events to -%s" % time.strftime("%H:%M:%S", time.gmtime(histseconds)))

		LANG_FLAG = 0x800    # bit 11 indicates UTF-8 for filenames
		with Timer() as timer:
			logger.info("Enumerating JTV event information")
			with zipfile.ZipFile(jtv_filename, 'r') as archive:
				for zinfo in filter(lambda x: x.filename.endswith('.pdt'), archive.infolist()):
					filename = zinfo.filename
					if zinfo.flag_bits & LANG_FLAG == 0:
						if PY3:
							filename = bytes(filename, encoding='cp437')
						unicode_fname = filename.decode('cp866', 'replace')
					else:
						unicode_fname = filename
					if PY3:
						channel_id = ensure_str('%s' % unicode_fname[0:-4]).replace('_', ' ').lower()
					else:
						channel_id = ensure_text('%s' % unicode_fname[0:-4]).replace('_', ' ').lower().encode('utf-8')

					if channel_id not in channelsDict:
						continue
					try:
						titles = parse_titles(archive.read(zinfo.filename), jtv_encoding)
					except Exception as e:
						logger.warn("JTVConverter failed to process titles in {}: {}".format(zinfo.filename, repr(e)))
						continue
					try:
						events_fname = zinfo.filename[:-4] + '.ndx'
						events = parse_events(archive.read(events_fname))
						events = deque(zip(events[:-1], events[1:]))
					except Exception as e:
						logger.warn("JTVConverter failed to process events in {}: {}".format(events_fname, repr(e)))
						continue

					events_tuple = tuple()	#tuple of tuples

					for title in titles:
						timer.counter += 1
						try:
							t = list(events.popleft())
							try:
								title, desc = [ensure_str(s.strip()) for s in PATTERN.split(title)]
							except:
								try:
									title, desc = [ensure_str(s.strip()) for s in title.split('.', 1)]

								except:
									title, desc = ensure_str(title.strip()), ''

							if t[0] <= timespan and now_timestamp_utc <= t[1] >= t[0]:
								t[1] -= t[0]
								if t[1] >= 86400:
									raise IndexError(t)
								l = [title, '', desc]
								l.append(LANG if isDreamOS else 0)
								events_tuple += (tuple(t+l),)
								# here we get a tuple of tuples ;) or None
								# return a None object to give up time to the reactor.
								# 0. start time (int)
								# 1. duration (int)
								# 2. event title (string)
								# 3. short description (string)
								# 4. extended description (string)
								# 5. event type (byte) or ISO639_2 three-letter abbreviation langguage code in DreamOS (string)
						except IndexError:
							 pass # Title has no end time or unacceptably large duration
						except Exception as e:
							logger.error("JTVConverter parsing JTV event error: %s" % e)
						finally:
							t = int(time.time()) - histseconds
							if t - now_timestamp_utc >= 10:
								logger.info("Processed: %s events" % timer.counter)
								now_timestamp_utc = t

					yield (list(channelsDict[channel_id]), events_tuple)

			timer.msg = 'Processed {} events in {} or ~{:.0f} per second'.format(timer.counter, time.strftime("%H:%M:%S", time.gmtime(timer.elapsed)), timer.counter/timer.elapsed)
