# -*- coding: utf-8 -*-
from __future__ import division
import re, calendar, time
from .. import isDreamOS, xml_unescape
from .EPGConfig import enumerateXML, Timer, histseconds, timespan
from .EPGImport import ISO639_associations
from ..config import E2m3u2bConfig
from ..modules.six import iteritems
from ..modules.six.moves import filter, map
from Components.Language import language
from collections import OrderedDict
from copy import copy

logger = E2m3u2bConfig.logger
LANG = ISO639_associations.get(language.getLanguage()[:2], 'eng')
TAGS = ['title', 'sub-title', 'desc']
PATTERN = re.compile(r'\s+(?=(?:[,.?!:;]))')	# formalizing pinning marks
XMLTVDateParser = lambda x: time.struct_time(tuple(map(int, (x[:4], x[4:6], x[6:8], x[8:10], x[10:], 0, -1, -1, 0))))


def get_event_data(elem):
	r = dict()
	try:
		for node in filter(lambda x: x.tag in TAGS, elem.iter()):
			lang = ISO639_associations.get(node.get('lang'), 'eng')
			d = r.get(lang, OrderedDict.fromkeys(TAGS,''))
			d[node.tag] = PATTERN.sub('', xml_unescape(node.text)) if node.text else ''
			r[lang] = d
			node.clear()
		return r

	except Exception as e:
		logger.error("XMLTVConverter get_event_data error: %s" % e)
		return None

def get_timestamp_utc(xmltv_date):
	try:
		if xmltv_date:
			return calendar.timegm(XMLTVDateParser(xmltv_date[:12])) - ((3600 * int(xmltv_date[-5:]) // 100) if len(xmltv_date)>14 else 0)
		else:
			raise
	except Exception as e:
		logger.error("XMLTVConverter get_time_utc error: %s" % e)
		return 0

def getDataTuple(elem, now_timestamp_utc):
	'''here we get a event tuple of tuples or None
	return a None object to give up time to the reactor.

	The structure of the returned tuple:
	0. start time (int)
	1. duration (int)
	2. event title (string)
	3. short description (string)
	4. extended description (string)
	5. event type (byte) or ISO639_2 three-letter abbreviation langguage code in DreamOS (string)'''
	event = tuple()
	try:
		t = [get_timestamp_utc(elem.get(x,'').strip()) for x in ('start', 'stop')]
		if all(t) and (t[0] <= timespan) and (now_timestamp_utc <= t[1] > t[0]):
			t[1] -= t[0]
			if t[1] >= 86400:
				raise ValueError(t)

			d = get_event_data(elem)

			if isDreamOS:
				for lang, v in iteritems(d):
					l = list(v.values())
					if not(l and l[0]):
						logger.debug("Skipping [%s] event! The empty value of 'title' attribute is not allowed" % lang)
						continue
					if l[1]: l[1] += '\n'
					event += (tuple(t + l + [lang]),)
			else:
				# On OE < 2.5 priority is given to the Enigma2 interface language
				l = list(d.get(LANG, d[list(d)[-1]]).values())
				if not(l and l[0]):
					raise KeyError(l)
				if l[1]: l[1] += '\n'
				event = (tuple(t + l + [0]),)

	except KeyError:
		logger.debug("Skipping event! The empty value of 'title' attribute is not allowed")
	except ValueError:
		logger.debug("Skipping event! An unacceptably large 'duration' attribute value")
	except Exception as e:
		logger.debug("XMLTVConverter getDataTuple error: %s" % e)
	finally:
		return event

def pairwise(it):
	c = next(it)
	for n in it:
		yield (c, n)
		c = copy(n)
	yield (c, None)


class XMLTVConverter(object):

	@classmethod
	def enumFile(cls, fileobj, channelsDict={}):
		'''
		Here we yield a tuple consisting of a list of services and a tuple of event tuples or None.
		Return a None object to give up time to the reactor.
		'''
		if not channelsDict:
			logger.warn("XMLTVConverter is nothing to enumerate. The channels lookup table is empty")
			return
		now_timestamp_utc = int(time.time() - histseconds)
		logger.info("Keep outdated EPG events to -%s" % time.strftime("%H:%M:%S", time.gmtime(histseconds)))
		with Timer() as timer:
			logger.info("Enumerating XMLTV event information")
			events_tuple, id = tuple(), None
			for elem, next_elem in pairwise(enumerateXML(fileobj, 'programme')):
				timer.counter += 1
				if not id:
					id = elem.get('channel', '')
					channel_id = xml_unescape(id.lower())
					if channel_id not in channelsDict:
						continue

				event = getDataTuple(elem, now_timestamp_utc)
				if next_elem is not None:
					next_id = next_elem.get('channel', '')
				if event:
					events_tuple += event
					if next_id == id:
						continue
					yield (list(channelsDict[channel_id]), events_tuple)
				elif not event and next_id == id:
					continue
				elif events_tuple:
					yield (list(channelsDict[channel_id]), events_tuple)
				else:
					yield None
				events_tuple, id = tuple(), None

			timer.msg = 'Processed {} events in {} or ~{:.0f} per second'.format(timer.counter, time.strftime("%H:%M:%S", time.gmtime(timer.elapsed)), timer.counter/timer.elapsed)
